from . import MarkdownFile
import re
from datetime import datetime

CHANGELOG_ENTRY_REGEX = '.*\[([0-9]*\.[0-9]*\.[0-9]*)\] - [0-9]{4}-[0-9]{1,2}-[0-9]+'

class ChangelogFile(MarkdownFile):

    def __init__(self, github_file, repo):
        self.changelog_entry_types = [ 'added', 'updated', 'removed' ]
        self._generate_interface_methods()
        super().__init__(github_file, repo)

    def _load_objects(self):
        string_lines = [line.decode('utf-8') for line in super()._load_objects()]
        self.header = []
        changelog_entries = []
        changelog_entry_block = []
        empty_previous_line = False
        inside_changelog_entry_block = False
        footer_start = None

        for i, line in enumerate(string_lines):
            # set on previous iteration
            if empty_previous_line:
                if re.search(CHANGELOG_ENTRY_REGEX, line):

                    # reached the end of the header (start of the changelog)
                    if not self.header:
                        self.header = string_lines[0:i]

                    # reached end of previous block (start of next changelog entry)
                    if changelog_entry_block:
                        changelog_entries.append(changelog_entry_block)

                    inside_changelog_entry_block = True
                    changelog_entry_block = []

                # if previous line was empty and it did _not_ match the changelog entry regex,
                # it means the end of the changelog entries was reached, and now it's reading
                # through the footer.
                else:
                    inside_changelog_entry_block = False
                    footer_start = i

            if inside_changelog_entry_block:
                changelog_entry_block.append(line)

            # used for the subsequent iteration to determine when we've gone through all
            # the changelog entries
            empty_previous_line = not line

        changelog_entries.append(changelog_entry_block)
        if footer_start:
            self.footer = string_lines[footer_start:len(string_lines) - 1]
        else:
            self.footer =  []

        return changelog_entries

    def _dump(self):
        blocks = []
        for o in self.objects:
            blocks.append('\n'.join(o))

        entries = self._format_new_changelog_entry()

        lines = ['\n'.join(self.header)]
        if entries:
            lines.append('\n'.join(entries))
            lines.append('')
        lines.append('\n'.join(blocks))
        lines.append('\n'.join(self.footer))

        return '\n'.join(lines)

    def _format_new_changelog_entry(self):
        lines = []
        for changelog_entry_type in self.changelog_entry_types:
            lines += self._format_list_entry(changelog_entry_type)

        if lines:
            header = f"## [{self.repo.new_version}] - {self._format_date()}"
            lines.insert(0, header)
        return lines

    def _format_list_entry(self, list_type):
        lines = []
        entries = getattr(self, f'_{list_type}')
        for entry, ticket in entries:
            if ticket:
                line = f'{entry} [{ticket}]'
            else:
                line = entry
            lines.append(f'- {line}')

        if lines:
            header = f'### {list_type.capitalize()}'
            lines.insert(0, header)
        return lines

    def _generate_interface_methods(self):
        for changelog_entry_type in self.changelog_entry_types:
            internal_store = f'_{changelog_entry_type}'

            def fn(self, entry, ticket=None, store=internal_store):
                getattr(self, store).append((entry, ticket))

            setattr(ChangelogFile, changelog_entry_type, fn)
            setattr(ChangelogFile, internal_store, [])

    def _format_date(self):
        return datetime.now().strftime('%Y-%m-%d')



import logging
import sys
import argparse
from .transformations import SearchAndReplace
from .config import Config
from .repo import Repo
from github import GithubException

logger = logging.getLogger('gordian')
ch = logging.StreamHandler()
logger.setLevel(level=logging.INFO)
formatter = logging.Formatter('[%(asctime)-15s] %(levelname)s %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)


class VerboseLogging(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        super().__init__(option_strings=option_strings,
                         dest=dest,
                         nargs=0,
                         const=True,
                         default=None,
                         required=False)

    def __call__(self, parser, namespace, values, option_string=None):
        logger.setLevel(level=logging.DEBUG)
        formatter = logging.Formatter('[%(asctime)-15s] %(module)s.%(funcName)s %(levelname)s %(message)s')
        logger.handlers[0].setFormatter(formatter)


def get_basic_parser():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument(
        '-c', '--config',
        required=False,
        default='config.yaml',
        dest='config_file',
        help='Config file path.'
    )
    parser.add_argument(
        '-g', '--github-api',
        required=False,
        dest='github_api',
        help='Github API URL'
    )
    parser.add_argument(
        '--pr',
        required=True,
        dest='pr_message',
        help='Pull request name.'
    )
    parser.add_argument(
        '-v', '--verbose',
        required=False,
        action=VerboseLogging,
        dest='verbose',
        help='Enable debug output'
    )
    parser.add_argument(
        '-d', '--dry-run',
        required=False,
        action='store_true',
        dest='dry_run',
        help='Enable dry run mode'
    )
    parser.add_argument(
        '-b', '--branch',
        required=False,
        dest='branch',
        help='Branch name to use'
    )
    parser.add_argument(
        '-t', '--target-branch',
        required=False,
        default='master',
        dest='target_branch',
        help='Target branch'
    )
    parser.add_argument(
        '-l','--labels',
        required=False,
        default=[],
        nargs='+',
        dest='pr_labels',
        help='List of space separated label names you wish to add to your pull request(s)'
    )
    bumpers = parser.add_mutually_exclusive_group(required=False)
    bumpers.add_argument(
            '-M', '--major',
            dest='semver_label',
            action='store_const',
            const='major',
            help='Bump the major version.'
        )
    bumpers.add_argument(
            '-m', '--minor',
            dest='semver_label',
            action='store_const',
            const='minor',
            help='Bump the minor version.'
        )
    bumpers.add_argument(
            '-p', '--patch',
            dest='semver_label',
            action='store_const',
            const='patch',
            help='Bump the patch version.'
        )
    pr_desc = parser.add_mutually_exclusive_group(required=False)
    pr_desc.add_argument(
            '--description',
            default='',
            dest='description',
            help='Description to be passed to the PR.'
        )
    pr_desc.add_argument(
            '--description-file',
            dest='description_file',
            help='Local file path for the description to be passed to the PR.'
        )
    parser.add_argument(
        '--force-changelog',
        required=False,
        dest='force_changelog',
        help='Fail if changelog does not exist or cannot be parsed'
    )
    return parser


def create_parser(args):
    parser = get_basic_parser()

    parser.add_argument(
        '-s', '--search',
        required=True,
        action='append',
        dest='search',
        help='The string to search for in config files.'
    )
    parser.add_argument(
        '-r', '--replace',
        required=True,
        action='append',
        dest='replace',
        help='The string that will replace instances of the searched string.'
    )

    args = parser.parse_args(args)
    if len(args.search) != len(args.replace):
        raise argparse.ArgumentTypeError('Number of search and replace arguments must be the same!')
    return args

def apply_transformations(args, transformations, pr_created_callback=None):
    config = Config(args.config_file)
    pr_description = get_pr_description(args)
    transform(args, transformations, config.get_data(), pr_description, pr_created_callback=pr_created_callback)

def get_pr_description(args):
    if args.description_file is not None:
        with open(args.description_file,'r') as fh:
            return fh.read()
    return args.description

def transform(args, transformations, repositories, pr_description, pr_created_callback):
    pull_request_urls = []
    for repo_name in repositories:
        logger.info(f'Processing repo: {repo_name}')
        repo = Repo(repo_name, github_api_url=args.github_api, branch=args.branch, semver_label=args.semver_label, target_branch=args.target_branch)
        for transformation in transformations:
            transformation(args, repo).run()
        if repo.dirty:
            repo.bump_version(args.dry_run)
            if not args.dry_run:
                try:
                    pull_request = repo.create_pr(args.pr_message, pr_description, args.target_branch, args.pr_labels)
                    pull_request_urls.append(pull_request.html_url)
                    if pr_created_callback is not None:
                        logger.debug(f'Calling post pr created callback with: {pull_request}, {repo.branch_name}')
                        pr_created_callback(repo_name, pull_request)
                    logger.info(f'PR created: {args.pr_message}. Branch: {repo.branch_name}. Labels: {args.pr_labels}')
                except GithubException as e:
                    logger.info(f'PR already exists for {repo.branch_name}. Error: {e}')

    if pull_request_urls:
        logger.info('Pull requests')
        [ logger.info(url) for url in pull_request_urls ]


def main():
    args = create_parser(sys.argv[1:])
    apply_transformations(args, [SearchAndReplace])

if __name__ == '__main__':
    main()



import yaml

class Config:

    def __init__(self, config_file):
        self.config_file = config_file

    def get_data(self):
        with open(self.config_file, 'r') as stream:
            data_loaded = yaml.safe_load(stream)
        return data_loaded


from github import Github
from github import GithubException
import datetime
import logging
import os
import time
from retry import retry
from gordian.files import *

logger = logging.getLogger(__name__)

BASE_URL = 'https://api.github.com'


class Repo:

    def __init__(self, repo_name, github_api_url=None, branch=None, github=None, files=None, semver_label=None, target_branch='master'):
        if github_api_url is None:
            self.github_api_url = BASE_URL
        else:
            self.github_api_url = github_api_url

        if github is not None:
            self._github = github
        else:
            if "GIT_TOKEN" in os.environ:
                logger.debug('Using git token')
                self._github = Github(base_url=self.github_api_url, login_or_token=os.environ['GIT_TOKEN'])
            else:
                logger.debug('Using git username and password')
                self._github = Github(base_url=self.github_api_url, login_or_token=os.environ['GIT_USERNAME'], password=os.environ['GIT_PASSWORD'])

        if files is None:
            files = []
        self.files = files

        self.repo_name = repo_name
        self._original_repo = self._github.get_repo(repo_name)

        self.version_file = None
        self.changelog_file = None
        self.branch_exists = False
        self.dirty = False
        self.new_version = None
        self.semver_label = semver_label
        self.target_branch = None
        self.set_target_branch(target_branch)

        if branch:
            self.branch_name = f"refs/heads/{branch}"
        else:
            self.branch_name = f"refs/heads/{datetime.datetime.now().strftime('%Y-%m-%d-%H%M%S.%f')}"

        logger.debug(f'Github api url: {self.github_api_url}')
        logger.debug(f'Repo name: {self.repo_name}')
        logger.debug(f'Target ref: {self.target_branch}')
        logger.debug(f'Branch name for this changes: {self.branch_name}')

    def set_target_branch(self, branch):
        if branch == self.target_branch:
            return

        self.target_branch = branch
        self.target_ref = f"refs/heads/{self.target_branch}"
        # Resetting the file cache when we change the branch
        self.files = []

    def get_objects(self, filename, klass=None):
        file = self.find_file(filename)

        if filename == 'CHANGELOG.md':
            return self.changelog

        if file is None:
            raise FileNotFoundError

        if klass:
            return klass(file, self)

        _, ext = os.path.splitext(filename)

        if ext in ('.yaml', '.yml'):
            return YamlFile(file, self)
        if ext == '.json':
            return JsonFile(file, self)
        if ext == '.md':
            return MarkdownFile(file, self)

    def get_files(self):
        if not self.files:
            logger.debug(f'Getting repo content')
            contents = self._original_repo.get_contents('', self.target_ref)
            while contents:
                file = contents.pop(0)
                if file.path == 'version':
                    self.version_file = file
                elif file.path == 'CHANGELOG.md':
                    self.changelog = ChangelogFile(file, self)
                elif file.type == 'dir':
                    contents.extend(self._original_repo.get_contents(file.path, self.target_ref))
                else:
                    self.files.append(file)

        self._get_new_version()

        return self.files

    def find_file(self, filename):
        for file in self.get_files():
            if file.path == filename:
                return file

    def _make_branch(self):
        logger.info('Forking repo...')
        self._forked_repo = self._original_repo.create_fork()
        branch = self._get_branch()
        logger.debug(f'Creating branch {self.branch_name}')

        try:
            ref = self._forked_repo.create_git_ref(ref=self.branch_name, sha=branch.commit.sha)
        except GithubException as e:
            logger.debug(f'Branch {self.branch_name} already exists in github')
        self.branch_exists = True

    @retry(GithubException, tries=3, delay=1, backoff=2)
    def _get_branch(self):
        logger.debug(f'Fetching branch {self.target_branch}...')
        return self._forked_repo.get_branch(self.target_branch)

    def bump_version(self, dry_run=False):
        if self.new_version is None:
            return

        logger.info(f'Bumping version {self.new_version}')
        self.update_file(
            self.version_file,
            self.new_version,
            f'Bumping version to {self.new_version}',
            dry_run
        )

    def update_file(self, repo_file, content, message, dry_run=False):
        self.dirty = True
        logger.info(message)
        if dry_run:
            logger.info('dry-run')
            return

        if not self.branch_exists:
            self._make_branch()

        logger.debug(f'Updating file {repo_file.path}')
        self._forked_repo.update_file(
            repo_file.path,
            message,
            content,
            repo_file.sha,
            branch=self.branch_name
        )

    def create_file(self, path, contents, message, dry_run=False):
        self.dirty = True
        logger.info(message)
        if dry_run:
            logger.info('dry-run')
            return

        if not self.branch_exists:
            self._make_branch()

        logger.debug(f'Creating file {path}')
        self._forked_repo.create_file(
            path,
            message,
            contents,
            branch=self.branch_name
        )

    def delete_file(self, file, message, dry_run=False):
        self.dirty = True
        logger.info(message)
        if dry_run:
            logger.info('dry-run')
            return

        if not self.branch_exists:
            self._make_branch()

        logger.debug(f'Deleting file {file.path}')
        self._forked_repo.delete_file(
            file.path,
            message,
            file.sha,
            branch=self.branch_name
        )

    def create_pr(self, pr_message, pr_body, target_branch, labels):
        pr = self._original_repo.create_pull(
            pr_message,
            pr_body,
            target_branch,
            f'{self._forked_repo.owner.login}:{self.branch_name}'
        )
        pr.set_labels(*labels)
        return pr

    def _get_new_version(self):
        if self.semver_label is None:
            return

        if self.version_file is None:
            logger.info('There is no version file in the repository, skipping bumping')
            return

        version = self.version_file.decoded_content.decode('utf-8')
        major, minor, patch = version.split('.')
        if self.semver_label == 'major':
            major = str(int(major) + 1)
            minor = patch = '0'
        elif self.semver_label == 'minor':
            minor = str(int(minor) + 1)
            patch = '0'
        elif self.semver_label == 'patch':
            patch = str(int(patch) + 1)
        self.new_version = '.'.join([major, minor, patch])



import logging
import yaml
import re
import jsonpatch
import copy
from deepdiff import DeepDiff

logger = logging.getLogger(__name__)


class Transformation(object):

    def __init__(self, args, repo):
        self.args = args
        self.repo = repo
        self.dry_run = args.dry_run

    def run(self):
        raise NotImplementedError('Please subclass the transformation and overrite this method')


class SearchAndReplace(Transformation):

    def __init__(self, args, repo):
        super().__init__(args, repo)
        self.changesets = zip(args.search, args.replace)

    def is_word_found(self, file, search):
        return search in str(file)

    def run(self):
        changes = False
        for search, replace in self.changesets:
            for file in self.repo.get_files():
                file_str = file.decoded_content
                if not self.is_word_found(file_str, search):
                    logger.debug(f"Ignoring {file}")
                    continue
                file_str = file_str.replace(search.encode(), replace.encode())
                changes = True

                message = f"Replacing '{search}' with '{replace}' in {file}"
                self.repo.update_file(file, file_str, message, self.dry_run)
        return changes


class JsonPatch(Transformation):

    def __init__(self, args, repo):
        super().__init__(args, repo)
        self.patch_path = args.patch_path
        self.file_regexp = args.file_regexp
        self.patch = jsonpatch.JsonPatch.from_string(open(self.patch_path, 'r').read())

    def run(self):
        changes = False
        for f in self.repo.get_files():
            file_changes = False
            if not re.match(self.file_regexp, f.path):
                continue
            logger.info(f'Path name: {f.path}')
            k8s_patches = list(yaml.safe_load_all(f.decoded_content))
            for r in k8s_patches:
                original = copy.deepcopy(r)
                try:
                    self.patch.apply(r, in_place=True)
                except jsonpatch.JsonPointerException as e:
                    logger.debug(e)
                except jsonpatch.JsonPatchTestFailed as e:
                    logger.debug(e)
                diff = DeepDiff(r, original)
                if diff != {}:
                    logger.info(f'Detected changes: {diff}')
                    message = f'Applied json patch defined in {self.patch_path}'
                    file_changes = True
            if file_changes:
                changes = True
                file_str = yaml.dump_all(k8s_patches, default_flow_style=False, explicit_start=True)
                logger.debug(file_str)
                self.repo.update_file(f, file_str, message, self.dry_run)
        return changes
